########
# Copyright (c) 2016 GigaSpaces Technologies Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
############

from je.configuration import configuration


class Work(object):

    def init(self):
        self.dir.mkdirs_p()

    def clear(self):
        for f in self.dir.walkfiles():
            if f.splitext()[1] == '.log':
                f.remove()
        # reversed so we remove children before parents
        dirs = reversed(list(self.dir.walkdirs()))
        for d in dirs:
            d.rmtree_p()

    @property
    def dir(self):
        return configuration.workdir

    def build_dir(self, job, build):
        result = self.dir / '{}-{}'.format(job, build)
        if not result.isdir():
            result.mkdir_p()
        return result

    def passed_dir(self, job, build):
        result = self.build_dir(job, build) / 'passed'
        if not result.isdir():
            result.mkdir_p()
        return result

    def failed_dir(self, job, build):
        result = self.build_dir(job, build) / 'failed'
        if not result.isdir():
            result.mkdir_p()
        return result

    def log_path(self, job, build):
        return self.build_dir(job, build) / 'console.log'
work = Work()



########
# Copyright (c) 2016 GigaSpaces Technologies Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
############

from je.jenkins import jenkins


class Completion(object):

    @staticmethod
    def job_completer(prefix, **kwargs):
        jobs = jenkins.list_jobs().get('jobs', [])
        for j in jobs:
            name = j.get('name')
            if name.startswith(prefix):
                yield name

    @staticmethod
    def build_completer(prefix, parsed_args, **kwargs):
        job = parsed_args.job
        builds = jenkins.list_builds(job, only_number=True)
        for build in builds:
            build = str(build)
            if build.startswith(prefix):
                yield build
completion = Completion()



########
# Copyright (c) 2016 GigaSpaces Technologies Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
############

import time

import argh
import requests

from je.configuration import configuration
from je.cache import cache


class Jenkins(object):

    def list_jobs(self):
        return self._query('', tree='jobs[name]')

    def build_job(self, job, parameters=None):
        resource = 'job/{}/buildWithParameters'.format(job)
        response = self._raw_query(resource=resource,
                                   method='POST',
                                   data=parameters)
        if response.status_code != 201:
            raise argh.CommandError(
                'Failed building job: {} [status={}, parameters={}]'
                .format(job, response.status_code, parameters))

    def list_builds(self, job, only_number=False):
        if only_number:
            tree = 'builds[number]'
        else:
            tree = 'builds[number,result,actions[causes[shortDescription]],' \
                   'timestamp,building]'
        builds = self._query('job/{}'.format(job), tree=tree)
        results = []
        for build in builds['builds']:
            number = build['number']
            if only_number:
                results.append(number)
                continue
            result = build['result']
            actions = build['actions']
            timestamp = build['timestamp']
            building = build['building']
            causes = []
            for action in actions:
                action_causes = action.get('causes')
                if not action_causes:
                    continue
                for cause in action_causes:
                    description = cause.get('shortDescription')
                    if not description:
                        continue
                    causes.append(description)
            results.append({
                'number': number,
                'result': result,
                'cause': ', '.join(causes),
                'timestamp': timestamp,
                'building': building
            })
        return reversed(results)

    def fetch_build(self, job, build):
        build_key = '{}-{}'.format(job, build)
        cached_build = cache.load(build_key)
        if cached_build:
            return cached_build
        print 'Build {} not in cache, ' \
              'retrieving from jenkins'.format(build_key)
        resource = 'job/{}/{}'.format(job, build)
        build = self._query(
            resource,
            tree='actions['
                 'causes[shortDescription],'
                 'parameters[name,value]'
                 '],'
                 'result,'
                 'duration,'
                 'timestamp,'
                 'building')
        building = build.get('building')
        if not building:
            try:
                test_report = self._query('{}/testReport'.format(resource))
            except Exception:
                test_report = {'status': 'error'}
        else:
            test_report = {}
        result = {
            'build': build,
            'test_report': test_report
        }
        if not building:
            cache.save(build_key, result)
        return result

    def fetch_build_logs(self, job, build):
        build_key = '{}-{}'.format(job, build)
        build_logs = cache.load_log(build_key)
        if build_logs:
            return build_logs
        print 'Logs not in cache, retrieving from jenkins'
        resource = 'job/{}/{}/consoleText'.format(job, build)
        response = self._raw_query(resource)
        build_logs = response.text
        fetched_build = self.fetch_build(job, build)
        if not fetched_build['build'].get('building'):
            cache.save_log(build_key, build_logs)
        return build_logs

    def tail_build_logs(self, job, build):
        build_key = '{}-{}'.format(job, build)
        build_logs = cache.load_log(build_key)
        if build_logs:
            yield build_logs
        else:
            resource = 'job/{}/{}/logText/progressiveText'.format(job, build)
            next_index = 0
            while True:
                response = self._raw_query(resource,
                                           method='POST',
                                           data={'start': next_index})
                previous_index = next_index
                next_index = int(response.headers.get('X-Text-Size'))
                if previous_index == next_index:
                    time.sleep(1)
                else:
                    yield response.text
                if response.headers.get('X-More-Data') != 'true':
                    break

    def _query(self, resource, tree=None):
        if tree:
            tree = 'tree={}'.format(tree)
        else:
            tree = ''
        resource = '{}/api/json?{}'.format(resource, tree)
        response = self._raw_query(resource)
        return response.json()

    @staticmethod
    def _raw_query(resource, method='GET', data=None):
        url = '{}/{}/{}'.format(configuration.jenkins_base_url,
                                configuration.jenkins_system_tests_base,
                                resource)
        response = requests.request(method, url,
                                    auth=(configuration.jenkins_username,
                                          configuration.jenkins_password),
                                    data=data)
        if response.status_code == 404:
            raise argh.CommandError('Resource not found. (404)'.format(
                resource))
        return response
jenkins = Jenkins()



########
# Copyright (c) 2016 GigaSpaces Technologies Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
############

import argparse
import datetime
import sys

import yaml
import colors
import argh
from argh.decorators import arg
from path import path

from je.jenkins import jenkins
from je.cache import cache
from je.configuration import configuration
from je.completion import completion
from je.work import work


app = argh.EntryPoint('je')
command = app


@command
@arg('--jenkins-username', required=True)
@arg('--jenkins-password', required=True)
@arg('--jenkins-base_url', required=True)
def init(jenkins_username=None,
         jenkins_password=None,
         jenkins_base_url=None,
         jenkins_system_tests_base=None,
         workdir=None,
         reset=False):
    configuration.save(jenkins_username=jenkins_username,
                       jenkins_password=jenkins_password,
                       jenkins_base_url=jenkins_base_url,
                       jenkins_system_tests_base=jenkins_system_tests_base,
                       workdir=workdir,
                       reset=reset)
    work.init()


@command
def list_jobs():
    jobs = jenkins.list_jobs()
    for job in jobs['jobs']:
        print job.get('name')


@command
@argh.named('list')
@arg('job', completer=completion.job_completer)
def ls(job):
    builds = jenkins.list_builds(job)
    for build in builds:
        result = build['result']
        building = build['building']
        number = str(build['number'])
        cause = build['cause']
        build_datetime = _timestamp_to_datetime(build['timestamp'])
        if building:
            build_color = colors.white
            result = 'BUILDING'
        elif result == 'FAILURE':
            build_color = colors.red
        elif result == 'ABORTED':
            build_color = colors.yellow
        else:
            build_color = colors.green
        print '{:<4}{:<18}{} ({})'.format(number,
                                          build_color(result),
                                          cause,
                                          build_datetime)


@command
@arg('job', completer=completion.job_completer)
@arg('build', completer=completion.build_completer)
def report(job, build, failed=False):
    build_number = build
    build = jenkins.fetch_build(job, build)
    report = build['test_report']
    build = build['build']
    if build.get('building'):
        return 'Building is currently running'
    if report.get('status') == 'error':
        return 'No tests report has been generated for this build'
    failed_dir = work.failed_dir(job, build_number)
    passed_dir = work.passed_dir(job, build_number)
    for d in [passed_dir, failed_dir]:
        d.rmtree_p()
        d.mkdir()
    build_parameters = _extract_build_parameters(build)
    interesting_parameters = ['system_tests_branch', 'system_tests_descriptor']
    interesting_parameters = {k: v for k, v in build_parameters.items()
                              if k in interesting_parameters}
    cause = _extract_build_cause(build)
    timestamp = build['timestamp']
    print '{}: {} ({})'.format(colors.bold('Cause'),
                               cause,
                               _timestamp_to_datetime(timestamp))
    print
    print colors.bold('Parameters: ')
    print yaml.safe_dump(interesting_parameters, default_flow_style=False)
    for suite in report['suites']:
        suite_name = suite['name']
        cases = []
        has_passed = False
        has_failed = False
        for case in suite['cases']:
            test_status = case['status']
            if test_status in ['FAILED', 'REGRESSION']:
                test_status = 'FAILED'
                colored_status = colors.red(test_status)
                has_failed = True
            elif test_status in ['PASSED', 'FIXED']:
                test_status = 'PASSED'
                colored_status = colors.green(test_status)
                has_passed = True
            elif test_status == 'SKIPPED':
                colored_status = colors.yellow(test_status)
                has_failed = True
            else:
                colored_status = test_status
            name = case['name']
            if not failed or test_status != 'PASSED':
                cases.append('{:<18}{}'.format(
                    colored_status,
                    name.split('@')[0]))
            filename = '{}.log'.format(name.replace(' ', '-'))
            dirname = passed_dir if test_status == 'PASSED' else failed_dir
            with open(dirname / filename, 'w') as f:
                f.write('name: {}\n\n'.format(case['name']))
                f.write('status: {}\n\n'.format(case['status']))
                f.write('class: {}\n\n'.format(case['className']))
                f.write('duration: {}\n\n'.format(case['duration']))
                f.write('error details: {}\n\n'.format(
                    case['errorDetails']))
                f.write('error stacktrace: {}\n\n'.format(
                    case['errorStackTrace']))
                f.write('stdout: \n{}\n\n'.format(case['stdout']))
                f.write('stderr: \b{}\n\n'.format(case['stderr']))
        if has_passed and has_failed:
            suite_name_color = colors.yellow
        elif has_passed:
            suite_name_color = colors.green
        elif has_failed:
            suite_name_color = colors.red
        else:
            suite_name_color = colors.white
        if cases:
            print suite_name_color(colors.bold(suite_name))
            print suite_name_color(colors.bold('-' * (len(suite_name))))
            print '\n'.join(cases)
            print
    print 'Report files written to {}'.format(work.build_dir(job,
                                                             build_number))


@command
@arg('job', completer=completion.job_completer)
@arg('builds',
     completer=completion.build_completer,
     nargs=argparse.ONE_OR_MORE)
def analyze(job, builds, passed_at_least_once=False, failed=False):
    build_numbers = set()
    for build in builds:
        split = build.split('-')
        if len(split) > 2:
            raise argh.CommandError('Illegal build range: {}'.format(build))
        elif len(split) == 1:
            build_numbers.add(build)
        else:
            start, stop = int(split[0]), int(split[1])
            build_numbers |= set(str(i) for i in range(start, stop+1))
    builds = [jenkins.fetch_build(job, b) for b in build_numbers]
    report = {}
    for build in builds:
        if build['build'].get('building'):
            print 'Skipping build {} as it currently running'.format(
                    build['number'])
            continue
        test_report = build['test_report']
        if test_report.get('status') == 'error':
            print 'Skipping build {} as no test reports were generated for it'\
                .format(build['number'])
        for suite in test_report['suites']:
            suite_name = suite['name']
            report_suite = report.get(suite_name, {})
            for case in suite['cases']:
                case_name = case['name'].split('@')[0]
                test_status = case['status']
                if test_status in ['FAILED', 'REGRESSION']:
                    test_status = 'FAILED'
                elif test_status in ['PASSED', 'FIXED']:
                    test_status = 'PASSED'
                report_case = report_suite.get(case_name, {})
                report_case_status = report_case.get(test_status, 0)
                report_case_status += 1
                report_case[test_status] = report_case_status
                report_suite[case_name] = report_case
            report[suite_name] = report_suite
    for suite_name, suite in report.items():
        cases = []
        suite_has_passed = False
        suite_has_failed = False
        for case_name, case in sorted(suite.items()):
            pass_count = case.get('PASSED', 0)
            fail_count = case.get('FAILED', 0)
            skip_count = case.get('SKIPPED', 0)
            case_has_failed = False
            if pass_count and (fail_count or skip_count) \
                    and not passed_at_least_once:
                case_color = colors.yellow
                suite_has_failed = case_has_failed = True
            elif pass_count:
                case_color = colors.green
                suite_has_passed = True
            else:
                case_color = colors.red
                suite_has_failed = case_has_failed = True
            if not failed or case_has_failed:
                cases.append('{} [passed={}, failed={}, skipped={}]'.
                             format(case_color(case_name),
                                    pass_count, fail_count, skip_count))
        if suite_has_passed and suite_has_failed:
            suite_name_color = colors.yellow
        elif suite_has_passed:
            suite_name_color = colors.green
        elif suite_has_failed:
            suite_name_color = colors.red
        else:
            suite_name_color = colors.white
        if cases:
            print suite_name_color(colors.bold(suite_name))
            print suite_name_color(colors.bold('-' * (len(suite_name))))
            print '\n'.join(cases)
            print


@command
@arg('job', completer=completion.job_completer)
@arg('build', completer=completion.build_completer)
def logs(job, build, stdout=False, tail=False):
    log_path = work.log_path(job, build)
    if not tail:
        result = jenkins.fetch_build_logs(job, build)
        if stdout:
            return result
        else:
            log_path.write_text(result, encoding='utf8')
            print 'Log file written to {}'.format(log_path)
    else:
        if stdout:
            stream = sys.stdout
        else:
            stream = open(log_path, 'w')
            print 'Log file written to {}'.format(log_path)
        try:
            for chunk in jenkins.tail_build_logs(job, build):
                stream.write(chunk.encode(encoding='utf8'))
                stream.flush()
        except KeyboardInterrupt:
            pass
        finally:
            if not stdout:
                stream.close()


@command
@arg('job', completer=completion.job_completer)
def build(job, branch=None, descriptor=None, source=None):
    parameters = {}
    if source:
        source_path = path(source).expanduser()
        if source_path.exists():
            parameters = yaml.safe_load(source_path.text())
        else:
            try:
                source = int(source)
            except ValueError:
                raise argh.CommandError('Invalid source: {}'.format(source))
            fetched_build = jenkins.fetch_build(job, source)
            parameters = _extract_build_parameters(fetched_build['build'])
    if branch:
        parameters['system_tests_branch'] = branch
    if descriptor:
        parameters['system_tests_descriptor'] = descriptor
    jenkins.build_job(job, parameters=parameters)
    print 'Build successfully queued [job={}, parameters={}]'.format(
        job, parameters)


@command
@arg('job', completer=completion.job_completer)
@arg('build', completer=completion.build_completer)
def parameters(job, build):
    build = jenkins.fetch_build(job, build)
    result = _extract_build_parameters(build['build'])
    return yaml.safe_dump(result, default_flow_style=False)


@command
def clear(force=False):
    if not force:
        raise argh.CommandError('clear will remove the cache directory and '
                                'clean the work diretory. pass --force if '
                                'this is indeed what you intend to do')
    else:
        cache.clear()
        work.clear()


@command
def workdir():
    return configuration.workdir


def _extract_build_parameters(build):
    actions = build['actions']
    for action in actions:
        action_parameters = action.get('parameters')
        if not action_parameters:
            continue
        if not any([parameter.get('name') == 'system_tests_branch'
                    for parameter in action_parameters]):
            continue
        return {p['name']: p['value'] for p in action_parameters}
    else:
        raise argh.CommandError('Invalid build {}'.format(build['build']))


def _extract_build_cause(build):
    causes = []
    actions = build['actions']
    for action in actions:
        action_causes = action.get('causes')
        if not action_causes:
            continue
        for cause in action_causes:
            description = cause.get('shortDescription')
            if not description:
                continue
            causes.append(description)
    return ', '.join(causes)


def _timestamp_to_datetime(timestamp):
    datetime_obj = datetime.datetime.fromtimestamp(timestamp / 1000.0)
    return datetime_obj.strftime('%Y-%m-%d %H:%M:%S')



########
# Copyright (c) 2016 GigaSpaces Technologies Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
############

import argh

import yaml
from path import path


class Configuration(object):

    def save(self,
             jenkins_username,
             jenkins_password,
             jenkins_base_url,
             jenkins_system_tests_base,
             workdir,
             reset):
        conf = self.conf_dir / 'config.yaml'
        if conf.exists() and not reset:
            raise argh.CommandError('Already initialized. '
                                    'Run "je init --reset"')
        if jenkins_base_url.endswith('/'):
            jenkins_base_url = jenkins_base_url[:-1]
        if not jenkins_system_tests_base:
            jenkins_system_tests_base = 'view/core_tests/job/dir_system-tests'
        workdir = workdir or self.conf_dir / 'work'
        workdir = path(workdir).expanduser().abspath()
        conf.write_text(yaml.safe_dump({
            'jenkins_username': jenkins_username,
            'jenkins_password': jenkins_password,
            'jenkins_base_url': jenkins_base_url,
            'jenkins_system_tests_base': jenkins_system_tests_base,
            'workdir': str(workdir)
        }, default_flow_style=False))

    @property
    def conf_dir(self):
        return path('~/.je').expanduser()

    @property
    def conf(self):
        conf = self.conf_dir / 'config.yaml'
        if not conf.exists():
            raise argh.CommandError('Not initialized. Run "je init"')
        return yaml.safe_load(conf.text())

    @property
    def jenkins_base_url(self):
        return self.conf.get('jenkins_base_url')

    @property
    def jenkins_username(self):
        return self.conf.get('jenkins_username')

    @property
    def jenkins_password(self):
        return self.conf.get('jenkins_password')

    @property
    def jenkins_system_tests_base(self):
        return self.conf.get('jenkins_system_tests_base')

    @property
    def workdir(self):
        return path(self.conf['workdir'])
configuration = Configuration()



########
# Copyright (c) 2016 GigaSpaces Technologies Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
############

import sys
from StringIO import StringIO

import argh

from je import commands


def main():
    parser = argh.ArghParser()
    subparsers_action = argh.utils.get_subparsers(parser, create=True)
    subparsers_action.metavar = ''
    parser.add_commands(commands.app.commands)
    errors = StringIO()
    parser.dispatch(errors_file=errors)
    errors_value = errors.getvalue()
    if errors_value:
        errors_value = errors_value.replace('CommandError', 'error').strip()
        sys.exit(errors_value)

if __name__ == '__main__':
    main()



########
# Copyright (c) 2016 GigaSpaces Technologies Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
############

import json

from je.configuration import configuration


class Cache(object):

    @property
    def cache_dir(self):
        return configuration.conf_dir / 'cache'

    @property
    def logs_dir(self):
        return self.cache_dir / 'logs'

    def save(self, key, value):
        key_path = self._key_path(key)
        key_path.write_text(json.dumps(value,
                                       indent=2,
                                       sort_keys=True))

    def save_log(self, key, value):
        log_path = self._log_path(key)
        log_path.write_text(value, encoding='utf8')

    def load(self, key):
        key_path = self._key_path(key)
        if not key_path.exists():
            return None
        return json.loads(key_path.text())

    def load_log(self, key):
        log_path = self._log_path(key)
        if not log_path.exists():
            return ''
        return log_path.text(encoding='utf8')

    def clear(self):
        self.cache_dir.rmtree_p()
        self.cache_dir.mkdir()
        self.logs_dir.mkdir()

    def _log_path(self, key):
        return self.logs_dir / '{}.log'.format(key)

    def _key_path(self, key):
        return self.cache_dir / '{}.json'.format(key)
cache = Cache()
